/*
 * r2300.c: R2000 and R3000 specific mmu/cache code.
 *
 * Copyright (C) 1996 David S. Miller (dm@engr.sgi.com)
 *
 * with a lot of changes to make this thing work for R3000s
 * Tx39XX R4k style caches added. HK
 * Copyright (C) 1998, 1999, 2000 Harald Koerfgen
 * Copyright (C) 1998 Gleb Raiko & Vladimir Roganov
 * Copyright (C) 2001, 2004  Maciej W. Rozycki
 */
/*
 * cache operations for the LX4xxx/5xxx
 * copy from c-r3k.c(change based on linux2.4 c-lexra.c, the cache parts.
 * MMU operations are move to tlb-r3k.c
 * Alan Liu
 * 2007-7-12
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/module.h>

#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/mmu_context.h>
#include <asm/system.h>
#include <asm/isadep.h>
#include <asm/io.h>
#include <asm/bootinfo.h>
#include <asm/cpu.h>
#include <asm/mipsregs.h>

static unsigned long icache_size, dcache_size;		/* Size in bytes */
static unsigned long icache_lsize, dcache_lsize;	/* Size in bytes */

extern void (*flush_cache_all)(void);
extern void (*__flush_cache_all)(void);
extern void (*flush_cache_mm)(struct mm_struct *mm);
extern void (*flush_cache_range)(struct vm_area_struct *vma, unsigned long start, unsigned long end);
extern void (*flush_cache_page)(struct vm_area_struct *vma, unsigned long page,
        unsigned long pfn);
extern void (*flush_icache_range)(unsigned long start, unsigned long end);

extern void (*flush_cache_sigtramp)(unsigned long addr);
extern void (*local_flush_data_cache_page)(void * addr);
extern void (*flush_data_cache_page)(unsigned long addr);
extern void (*flush_icache_all)(void);


#undef DEBUG_CACHE

static void __init lx_probe_cache(void)
{
	dcache_size = 1024*8;
	icache_size = 1024*8;

	dcache_lsize = 32;/* cache line size */
	icache_lsize = 32;/* cache line size */
}

/**
 * This function is used to flush a region of the instruction cache. Because
 * Lexra processors do not accomodate instruction cache flushing using 
 * traditional MIPS methods, this function invalidates the entire instruction 
 * cache. This is accomplished using the Lexra specific IINVAL bit of the 
 * coprocessor 0 cache control register.
 * @param start The start of the memory range to flush. Ignored.
 * @param size The size in bytes of the memory range to flush. Ignored.
 */
static void __lx_flush_icache_range(unsigned long start,
				    unsigned long end)
{
	unsigned int controlReg;
	unsigned long flags;

	/*
	 * Deprecated, use local_irq_save
	 * Alan Liu
	 */
	//save_and_cli(flags);
	local_irq_save(flags);

	//controlReg = read_32bit_cp0_register(CP0_CCTL);
	controlReg = read_c0_cctl();

	//write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_IINVAL));
	//write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_IINVAL));
	write_c0_cctl(controlReg & ~CCTL_IINVAL);
	write_c0_cctl(controlReg | CCTL_IINVAL);

	/* delay to allow cache to be flushed */
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t" "nop\n\t" ".set\treorder\n\t");

	//write_32bit_cp0_register(CP0_CCTL, controlReg);
	write_c0_cctl(controlReg);
	
	/*
	 * Deprecated, use local_irq_restore
	 * Alan Liu
	 */
	//restore_flags(flags);
	local_irq_restore(flags);
}

/**
 * This function is used to flush a region of the data cache. To accomplish this
 * with a Lexra processor, we read from the corresponding uncached space in
 * KSEG1.
 * @param start The start of the memory range to flush.
 * @param size The size in bytes of the memory range to flush.
 */
static void __lx_flush_dcache_range(unsigned long start,
				    unsigned long size)
{
	int index;
	register volatile int *ptr;
	register int dummy;

	unsigned long flags;
	size = (size + 8) >> 2;	/* bytes to words rounded to nearest word */

	/* Don't have to do anything to KSEG1 addresses */
	if (start == (unsigned long) KSEG1ADDR(start))
		return;

	//save_and_cli(flags);
	local_irq_save(flags);

	ptr = (volatile int *) KSEG1ADDR(start & 0xfffffffc);
	for (index = 0; index < size; index++)
		dummy += *ptr++;

	//restore_flags(flags);
	local_irq_restore(flags);
}

/*
 * It seems this function is not used...
 * Alan Liu
 * 2007/07/13
 */
#if 0
static inline unsigned long get_phys_page (unsigned long addr,
					   struct mm_struct *mm)
{
	pgd_t *pgd;
	pud_t *pud;
	pmd_t *pmd;
	pte_t *pte;
	unsigned long physpage;

	pgd = pgd_offset(mm, addr);
	pud = pud_offset(pgd, addr);
	pmd = pmd_offset(pud, addr);
	pte = pte_offset(pmd, addr);

	if ((physpage = pte_val(*pte)) & _PAGE_VALID)
		return KSEG0ADDR(physpage & PAGE_MASK);

	return 0;
}
#endif

/**
 *
 */
//static inline void lx_flush_cache_all(void)
static inline void lx_flush_cache_all(void)
{
	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

static inline void lx___flush_cache_all(void)
{
	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/*
 * This function is used to flush a portion of the instruction cache specified
 * by a mm_struct. For the Lexra processor we can only flush the entire cache 
 * so this is done by calling lx_flush_cache_range() and specifying the whole 
 * of KSEG0 as the range.
 * Why r3k_flush_cache_range is blank???	2007/07/13
 * Alan Liu
 */
//static void lx_flush_cache_mm(struct vm_area_struct *vma,
//unsigned long start, unsigned long end)
static void lx_flush_cache_mm(struct mm_struct *mm)
{
	if(!(cpu_context(smp_processor_id(), mm)))
		return;
	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/**
 * This function is used to flush a portion of the instruction cache specified
 * by a mm_struct. For the Lexra processor we can only flush the entire cache 
 * so this is done by calling lx_flush_cache_range() and specifying the whole 
 * of KSEG0 as the range.
 * @param mm A mm_struct pointer.
 * @param start Start of memory region to flush.
 * @param end End of memory region to flush.
 */
static void lx_flush_cache_range(struct vm_area_struct *vma,
				 unsigned long start, unsigned long end)
{
	if(!(cpu_context(smp_processor_id(), vma->vm_mm)))
		return;
	/* we are always going to just flush the entire cache for now */
	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/**
 * This function is used to flush a portion of the instruction cache specified
 * by a vm_area_struct. For the Lexra processor we can only flush the entire 
 * cache so this is done by calling lx_flush_cache_range() and specifying the 
 * whole of KSEG0 as the range.
 * @param vma A vm_area_struct pointer.
 * @param page The page to flush.
 */
static void lx_flush_cache_page(struct vm_area_struct *vma,
				unsigned long page)
{
	if(!(cpu_context(smp_processor_id(), vma->vm_mm)))
		return;
	__lx_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

static void local_lx_flush_data_cache_page(void *addr)
{
}

static void lx_flush_data_cache_page(unsigned long addr)
{
}

/**
 * Traditional R3000 processors flush the first three
 * words ot the cache starting at addr. We must flush
 * the entire cache.
 * @param addr Variable specifying the beginning of the  flush.
 */
static void lx_flush_cache_sigtramp(unsigned long addr)
{
	/**
	 * Traditional R3000 processors flush the first three
	 * words ot the cache starting at addr. We must flush
	 * the entire cache.
	 */
	__lx_flush_icache_range(addr, KSEG0 + icache_size);
}

/**
 * This function is used to flush and invalidate the memory used for DMA 
 * accesses.  Because Lexra uses a write through cache, it is only necessary 
 * to flush the specified range of the data cache.
 * @param start The start of the memory range used in the DMA transfer.
 * @param size The size in bytes of the DMA transfer.
 */
static void __lx_dma_cache_wback_inv(unsigned long start,
				     unsigned long size)
{
	__lx_flush_dcache_range(start, size);
}

void __init lx_cache_init(void)
{
	extern void build_clear_page(void);
	extern void build_copy_page(void);

	lx_probe_cache();

	flush_cache_all = lx_flush_cache_all;
	__flush_cache_all = lx___flush_cache_all;
	flush_cache_mm = lx_flush_cache_mm;
	flush_cache_range = lx_flush_cache_range;
	flush_cache_page = lx_flush_cache_page;
	flush_icache_range = __lx_flush_icache_range;

	flush_cache_sigtramp = lx_flush_cache_sigtramp;
	local_flush_data_cache_page = local_lx_flush_data_cache_page;
	flush_data_cache_page = lx_flush_data_cache_page;

	_dma_cache_wback_inv = __lx_dma_cache_wback_inv;
	_dma_cache_wback = __lx_dma_cache_wback_inv;
	_dma_cache_inv = __lx_dma_cache_wback_inv;

	printk("Primary instruction cache %ldkB, linesize %ld bytes.\n",
		icache_size >> 10, icache_lsize);
	printk("Primary data cache %ldkB, linesize %ld bytes.\n",
		dcache_size >> 10, dcache_lsize);

	/*
	 * I dont why use these two functions...
	 * Alan Liu
	 * 2007-7-13
	 */
	build_clear_page();
	build_copy_page();
}

EXPORT_SYMBOL(lx_flush_cache_all);
#if 0
/**
 * This function initializes all of the Lexra memory management functions.
 * This involves mapping the kernel cache management function pointers
 * to the Lexra specific functions, as well as setting the icache and
 * dcache size variables.
 */
void __init ld_mmu_lx(void)
{
	_clear_page = lx_clear_page;
	_copy_page = lx_copy_page;

	/*
	 * These variables are unused on this architecture.
	 */
	icache_size = 1024 * 8;
	dcache_size = 1024 * 8;

	_flush_cache_all = lx_flush_cache_all;
	___flush_cache_all = lx_flush_cache_all; 
	_flush_cache_mm = lx_flush_cache_mm;
	_flush_cache_range = lx_flush_cache_range;
	_flush_cache_page = lx_flush_cache_page;
	_flush_cache_sigtramp = lx_flush_cache_sigtramp;
	_flush_page_to_ram = lx_flush_page_to_ram;
	_flush_icache_page = lx_flush_icache_page;

	_flush_icache_range = __lx_flush_icache_range;
	// _flush_dcache_range = __lx_flush_dcache_range;

	_dma_cache_wback_inv = __lx_dma_cache_wback_inv;

	probe_tlb();

	_flush_cache_all();
	flush_tlb_all();

#ifdef CONFIG_MIPS_TRIHIDTV_PRO_AV_MIPS
	/*
	 * Try to protect the memory region of Host MIPS
	 * Or else maybe AV MIPS would crash the HOST KERNEL
	 * This is for the UTLB/BEV TLBS/TLBL
	 * It is in tlbex-r3k.S.
	 * Alan Liu
	 */
	memcpy((void *)bev0_except_entry, &except_vec0_r2300, 0x80);
	flush_icache_range(bev0_except_entry, bev0_except_entry + 0x80);
#else
	memcpy((void *)KSEG0, &except_vec0_r2300, 0x80);
	flush_icache_range(KSEG0, KSEG0 + 0x80);
#endif
}

/**
 * Assebly language function to clear a page of memory.
 * @param page Pointer to the page.
 */
void lx_clear_page(void *page)
{
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     ".set\tnoat\n\t"
			     "addiu\t$1,%0,%2\n"
			     "1:\tsw\t$0,(%0)\n\t"
			     "sw\t$0,4(%0)\n\t"
			     "sw\t$0,8(%0)\n\t"
			     "sw\t$0,12(%0)\n\t"
			     "addiu\t%0,32\n\t"
			     "sw\t$0,-16(%0)\n\t"
			     "sw\t$0,-12(%0)\n\t"
			     "sw\t$0,-8(%0)\n\t"
			     "bne\t$1,%0,1b\n\t"
			     "sw\t$0,-4(%0)\n\t"
			     ".set\tat\n\t"
			     ".set\treorder":"=r"(page):"0"(page),
			     "I"(PAGE_SIZE):"$1", "memory");
}


/**
 * This function is used to copy a page (4096 bytes) from the location
 * pointed to by 'from' to the location pointed to by 'to'.
 * @param to Location to copy to.
 * @param from Location to copy from.
 */
//static void lx_copy_page(void *to, void *from)
void lx_copy_page(void *to, void *from)
{
	unsigned long dummy1, dummy2;
	unsigned long reg1, reg2, reg3, reg4;

	/* This function copies a 4 kb page from addr 'from' to the addr
	 * pointed to by 'to'.
	 */
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     ".set\tnoat\n\t"
			     "addiu\t$1,%0,%8\n"
			     "1:\tlw\t%2,(%1)\n\t"
			     "lw\t%3,4(%1)\n\t"
			     "lw\t%4,8(%1)\n\t"
			     "lw\t%5,12(%1)\n\t"
			     "sw\t%2,(%0)\n\t"
			     "sw\t%3,4(%0)\n\t"
			     "sw\t%4,8(%0)\n\t"
			     "sw\t%5,12(%0)\n\t"
			     "lw\t%2,16(%1)\n\t"
			     "lw\t%3,20(%1)\n\t"
			     "lw\t%4,24(%1)\n\t"
			     "lw\t%5,28(%1)\n\t"
			     "sw\t%2,16(%0)\n\t"
			     "sw\t%3,20(%0)\n\t"
			     "sw\t%4,24(%0)\n\t"
			     "sw\t%5,28(%0)\n\t"
			     "addiu\t%0,64\n\t"
			     "addiu\t%1,64\n\t"
			     "lw\t%2,-32(%1)\n\t"
			     "lw\t%3,-28(%1)\n\t"
			     "lw\t%4,-24(%1)\n\t"
			     "lw\t%5,-20(%1)\n\t"
			     "sw\t%2,-32(%0)\n\t"
			     "sw\t%3,-28(%0)\n\t"
			     "sw\t%4,-24(%0)\n\t"
			     "sw\t%5,-20(%0)\n\t"
			     "lw\t%2,-16(%1)\n\t"
			     "lw\t%3,-12(%1)\n\t"
			     "lw\t%4,-8(%1)\n\t"
			     "lw\t%5,-4(%1)\n\t"
			     "sw\t%2,-16(%0)\n\t"
			     "sw\t%3,-12(%0)\n\t"
			     "sw\t%4,-8(%0)\n\t"
			     "bne\t$1,%0,1b\n\t"
			     "sw\t%5,-4(%0)\n\t"
			     ".set\tat\n\t"
			     ".set\treorder":"=r"(dummy1), "=r"(dummy2),
			     "=&r"(reg1), "=&r"(reg2), "=&r"(reg3),
			     "=&r"(reg4):"0"(to), "1"(from),
			     "I"(PAGE_SIZE));
}
#endif
