/*
 * MTD SPI driver for Winbond W25Qxx flash chips
 *
 * Author: Sam Sheng, samsheng@trident.com.cn
 *
 * Copyright (c) 2008, Trident Multimedia Technology Inc.
 *
 * Some parts are based on M25P80.c by Mike Lavender
 * Some parts are based on lart.c by Abraham Van Der Merwe
 *
 * Cleaned up and generalized based on mtd_dataflash.c
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/interrupt.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/spi/spi.h>
#include <linux/spi/flash.h>
#include <linux/delay.h>

#include <asm/semaphore.h>
#include <asm/trihidtv/trihidtv.h>


/* NOTE: AT 25F and SST 25LF series are very similar,
 * but commands for sector erase and chip id differ...
 */

#define FLASH_PAGESIZE		256


#define SPI_CMD_DUMMY				0x00
#define SPI_CMD_WRITE_ENABLE		0x06
#define SPI_CMD_WRITE_DISABLE		0x04
#define SPI_CMD_READ_DATA		0x03
#define SPI_CMD_FAST_READ_DATA		0x0b
#define SPI_CMD_READ_STATUS1		0x05
#define SPI_CMD_READ_STATUS2		0x35
#define SPI_CMD_WRITE_STATUS		0x01
#define SPI_CMD_PAGE_PROGRAM		0x02
#define SPI_CMD_QPAGE_PROGRAM		0x32
#define SPI_CMD_BLOCK_ERASE1		0xd8
#define SPI_CMD_BLOCK_ERASE2		0x52
#define SPI_CMD_SECTOR_ERASE		0x20
#define SPI_CMD_CHIP_ERASE			0xc7
#define SPI_CMD_ERASE_SUSPEND		0x75
#define SPI_CMD_ERASE_RESUME		0x7a
#define SPI_CMD_POWER_DOWN			0xb9
#define SPI_CMD_HP_MODE				0xa3
#define SPI_CMD_MODE_BR				0xff
#define SPI_CMD_RLS_PWR_DOWN		0xab
#define SPI_CMD_RLS_HPM				0xab
#define SPI_CMD_RLS_DID				0xab
#define SPI_CMD_MFR_ID				0x90
#define SPI_CMD_UNIQUE_ID			0x4b
#define SPI_CMD_JEDEC_ID			0x9f

#define SPI_REG_CNTL		0x1b002014
#define SPI_REG_COMMAND		0x1b002018
#define SPI_REG_DATA		0x1b00201c
#define SPI_REG_CLK_CNTL	0x1b002020
#define SPI_REG_TIME_CNTL0	0x1b002024
#define SPI_REG_TIME_CNTL1	0x1b002028

#define SPI_DATA_SET_SHIFT			0
#define SPI_DATA_GET_SHIFT			16
#define SPI_DATA_SET_MASK			(0xff << SPI_DATA_SET_SHIFT)
#define SPI_DATA_GET_MASK			(0xff << SPI_DATA_GET_SHIFT)
#define SPI_DATA_SET_STATUS_MASK	(0x1 << 24)
#define SPI_DATA_GET_STATUS_MASK	(0x1 << 25)

/* Define max times to check status register before we give up. */
#define SPI_RETRY_MAX		100000


#ifdef CONFIG_MTD_PARTITIONS
#define	mtd_has_partitions()	(1)
#else
#define	mtd_has_partitions()	(0)
#endif

/****************************************************************************/

struct w25q {
	struct spi_device	*spi;
	struct semaphore	lock;
	struct mtd_info		mtd;
	unsigned		partitioned;
	u8			command[4];
};

static inline struct w25q *mtd_to_w25q(struct mtd_info *mtd)
{
	return container_of(mtd, struct w25q, mtd);
}

/****************************************************************************/

/*
 * Internal helper functions
 */

static inline void
spi_set_data(unsigned int val)
{
	unsigned int t;
	
	t = ReadRegWord(SPI_REG_DATA);
	t &= ~(SPI_DATA_SET_MASK | SPI_DATA_GET_STATUS_MASK | SPI_DATA_SET_STATUS_MASK);
	t |= val;
	WriteRegWord(SPI_REG_DATA, t);
}

static inline void
spi_get_data(unsigned int *val)
{
	unsigned int t;
	
	t = ReadRegWord(SPI_REG_DATA);
	*val = ((t & SPI_DATA_GET_MASK) >> SPI_DATA_GET_SHIFT);
}

static inline int
spi_check_set_status(void)
{
	int retry = 0;
		
	while(retry++ < SPI_RETRY_MAX)
	{
		if(ReadRegWord(SPI_REG_DATA) & SPI_DATA_SET_STATUS_MASK) // bit 24
			return 1;
	}
	
	return 0;
}

static inline int
spi_check_get_status(void)
{
	int retry = 0;
	
	while(retry++ < SPI_RETRY_MAX)
	{
		if(ReadRegWord(SPI_REG_DATA) & SPI_DATA_GET_STATUS_MASK) // bit 25
		{
			return 1;
		}
	}
	
	return 0;
}

static inline void
spi_issue_write_cmd(void)
{
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031003);
	ReadRegWord(SPI_REG_COMMAND);
}

static inline void
spi_issue_read_cmd(void)
{
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031005);
	ReadRegWord(SPI_REG_COMMAND);
}

/*
 * Set write enable latch with Write Enable command.
 * Returns negative if error occurred.
 */
static inline int
spi_write_enable(void)
{
	int ret;

	WriteRegWord(SPI_REG_CNTL, 0x00000000);
	
	spi_set_data(SPI_CMD_WRITE_ENABLE);
	
	//spi_issue_write_cmd();
	WriteRegWord(SPI_REG_COMMAND, 0x04031001);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031003);
	ReadRegWord(SPI_REG_COMMAND);

	ret = spi_check_set_status();

	WriteRegWord(SPI_REG_COMMAND, 0x04031020);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	
	return ret;
}


/*
 * Set write disable latch with Write Disable command.
 * Returns negative if error occurred.
 */
static inline int
spi_write_disable(void)
{
	int ret;

	WriteRegWord(SPI_REG_CNTL, 0x00000000);
	
	spi_set_data(SPI_CMD_WRITE_DISABLE);
	
	//spi_issue_write_cmd();
	WriteRegWord(SPI_REG_COMMAND, 0x04031001);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031003);
	ReadRegWord(SPI_REG_COMMAND);

	ret = spi_check_set_status();

	WriteRegWord(SPI_REG_COMMAND, 0x04031020);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	
	return ret;
}

static inline unsigned int
clear_spi_cache(void)
{
	unsigned int data = 0;
	volatile unsigned int *addr = (unsigned int *)0xbc000000;

	__asm__ __volatile__(".set\tnoreorder\n\t.set\treorder"::: "memory");

	data += *addr;
	udelay(16);

	return data;
}

/*
 * Service routine to read status register until ready, or timeout occurs.
 * Returns non-zero if error.
 */
static int wait_till_ready(struct w25q *flash)
{
	int busy = 1;
	unsigned long timeout;

	/* The 500 msec is enough */
	timeout = jiffies + msecs_to_jiffies(500);

	udelay(16);

	while(time_before(jiffies, timeout))
	{
		unsigned int t;
		
		//spi_write_enable();
		
		// write total size
		WriteRegWord(SPI_REG_CNTL, 0x1);
		// write read length = 1
		WriteRegWord(SPI_REG_TIME_CNTL1, 0x13110501);
	
		WriteRegWord(SPI_REG_COMMAND, 0x04031000);
		ReadRegWord(SPI_REG_COMMAND);
		WriteRegWord(SPI_REG_COMMAND, 0x04031001);
		ReadRegWord(SPI_REG_COMMAND);
	
		//write cmd
		t = ReadRegWord(SPI_REG_DATA);
		t &= 0xfcffff00;
		t |= 0x5;
		WriteRegWord(SPI_REG_DATA, t);
		WriteRegWord(SPI_REG_COMMAND, 0x04031000);
		ReadRegWord(SPI_REG_COMMAND);
		WriteRegWord(SPI_REG_COMMAND, 0x04031003);
		ReadRegWord(SPI_REG_COMMAND);
		while(1)
		{
			t = ReadRegWord(SPI_REG_DATA);
			if(t & 0x01000000) // bit 24
				break;
		}
	
		//read status register
		WriteRegWord(SPI_REG_COMMAND, 0x04031000);
		ReadRegWord(SPI_REG_COMMAND);
		WriteRegWord(SPI_REG_COMMAND, 0x04031005);
		ReadRegWord(SPI_REG_COMMAND);
		while(1)
		{
			t = ReadRegWord(SPI_REG_DATA);
			if(t & 0x02000000) // bit 25
				break;
		}
	
		WriteRegWord(SPI_REG_COMMAND, 0x04031000);
		ReadRegWord(SPI_REG_COMMAND);
		WriteRegWord(SPI_REG_COMMAND, 0x04031020);
		ReadRegWord(SPI_REG_COMMAND);
		WriteRegWord(SPI_REG_COMMAND, 0x04031000);
		ReadRegWord(SPI_REG_COMMAND);
	
		spi_write_disable();
	
		WriteRegWord(SPI_REG_COMMAND, 0x04031101);
		ReadRegWord(SPI_REG_COMMAND);
	
		t = ((t & 0x00ff0000) >> 16);
		
		busy = (t & 0x1);
		if(!busy)
			break;

		schedule();
	}
	
	return !busy;
}


/*
 * Erase one sector of flash memory at offset ``offset'' which is any
 * address within the sector which should be erased.
 *
 * Returns 0 if successful, non-zero otherwise.
 */
static int erase_sector(struct w25q *flash, u32 offset)
{
	int i;
	unsigned char cmdset[4] = {SPI_CMD_BLOCK_ERASE1, SPI_CMD_DUMMY, 
		SPI_CMD_DUMMY, SPI_CMD_DUMMY};
	unsigned int ret = 0;

	DEBUG(MTD_DEBUG_LEVEL2, "%s: %s at 0x%08x\n", flash->mtd.name,
			__FUNCTION__, offset);
	
	spi_write_enable();

	cmdset[1] = ((offset >> 16) & 0xff);
	cmdset[2] = ((offset >> 8) & 0xff);
	cmdset[3] = ((offset >> 0) & 0xff);

	// write total size
	WriteRegWord(SPI_REG_CNTL, 0x3);
	// write read length = 0
	WriteRegWord(SPI_REG_TIME_CNTL1, 0x13010501);

	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031001);
	ReadRegWord(SPI_REG_COMMAND);

	for(i = 0; i < sizeof(cmdset) / sizeof(cmdset[0]); i++)
	{
		spi_set_data(cmdset[i]);

		spi_issue_write_cmd();
		
		if(!spi_check_set_status())
		{
			ret = -EIO;
			break;
		}
	}

	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031020);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031101);
	ReadRegWord(SPI_REG_COMMAND);

	if(!wait_till_ready(flash))
		ret = -EIO;

	clear_spi_cache();
	
	return ret;
}

/****************************************************************************/

/*
 * MTD implementation
 */

/*
 * Erase an address range on the flash chip.  The address range may extend
 * one or more erase sectors.  Return an error is there is a problem erasing.
 */
static int w25q16_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	struct w25q *flash = mtd_to_w25q(mtd);
	u32 addr,len;

	DEBUG(MTD_DEBUG_LEVEL2, "%s: %s %s 0x%08x, len %d\n",
			flash->mtd.name, __FUNCTION__, "at",
			(u32)instr->addr, instr->len);

	/* sanity checks */
	if (instr->addr + instr->len > flash->mtd.size)
		return -EINVAL;
	if ((instr->addr % mtd->erasesize) != 0
			|| (instr->len % mtd->erasesize) != 0) {
		return -EINVAL;
	}

	addr = instr->addr;
	len = instr->len;

  	down(&flash->lock);

	/* now erase those sectors */
	while (len) {
		if (erase_sector(flash, addr)) {
			instr->state = MTD_ERASE_FAILED;
			up(&flash->lock);
			return -EIO;
		}

		addr += mtd->erasesize;
		len -= mtd->erasesize;
	}

  	up(&flash->lock);

	instr->state = MTD_ERASE_DONE;
	mtd_erase_callback(instr);

	return 0;
}

/*
 * Read an address range from the flash chip.  The address range
 * may be any size provided it is within the physical boundaries.
 */
static int w25q16_read(struct mtd_info *mtd, loff_t from, size_t len,
	size_t *retlen, u_char *buf)
{
	struct w25q *flash = mtd_to_w25q(mtd);

	DEBUG(MTD_DEBUG_LEVEL2, "%s: %s %s 0x%08x, len %zd\n",
			flash->mtd.name, __FUNCTION__, "from",
			(u32)from, len);
	
	/* Byte count starts at zero. */
	if (retlen)
		*retlen = 0;
	
	/* sanity checks */
	if (!len)
		return 0;

	if (from + len > flash->mtd.size)
		return -EINVAL;

	down(&flash->lock);

	memcpy(buf, (unsigned char *)((unsigned char *)0xbc000000 + from), len);
	*retlen = len;

  	up(&flash->lock);

	return 0;
}

#ifdef SPI_WRITE_VERIFY
static void
buffer_dump(const unsigned char *buff, int len)
{
	int i;
	for(i = 0; i < len; i++)
	{
		if(i % 16 == 0)
			printk("\n");
		printk("0x%02X ", buff[i]);
	}
	printk("\n");
}
#endif

static int w25q16_write_page(struct mtd_info *mtd, loff_t to, size_t len,
	const u_char *buf)
{
	struct w25q *flash = mtd_to_w25q(mtd);
	int i;
	unsigned char cmdset[4] = {SPI_CMD_PAGE_PROGRAM, SPI_CMD_DUMMY, 
		SPI_CMD_DUMMY, SPI_CMD_DUMMY};

	cmdset[1] = ((to >> 16) & 0xff);
	cmdset[2] = ((to >> 8) & 0xff);
	cmdset[3] = ((to >> 0) & 0xff);

	udelay(16);
	spi_write_enable();

	// write total size
	WriteRegWord(SPI_REG_CNTL, len + 3);
	// write read length = 0
	WriteRegWord(SPI_REG_TIME_CNTL1, 0x13010501);

	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031001);
	ReadRegWord(SPI_REG_COMMAND);

	for(i = 0; i < sizeof(cmdset) / sizeof(cmdset[0]); i++)
	{
		spi_set_data(cmdset[i]);

		spi_issue_write_cmd();
		
		if(!spi_check_set_status())
		{
			return -EIO;
		}
	}

	for(i = 0; i < len; i++)
	{
		spi_set_data(buf[i]);

		spi_issue_write_cmd();
		
		if(!spi_check_set_status())
		{
			return -EIO;
		}
	}

	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031020);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031101);
	ReadRegWord(SPI_REG_COMMAND);

	if(!wait_till_ready(flash))
		return -EIO;

#ifdef SPI_WRITE_VERIFY
	clear_spi_cache();
	if(memcmp(buf, (void *)((char *)0xbc000000 + to), len))
	{
		printk("Write error at 0x%llx len 0x%x.\n", to, len);
		buffer_dump(buf, len);
		buffer_dump((unsigned char *)((char *)0xbc000000 + to), len);
	}
	else
	{
		/*printk("Write success at 0x%llx len 0x%x.\n", to, len);*/
	}
#endif
	
	return 0;
}

/*
 * Write an address range to the flash chip.  Data must be written in
 * FLASH_PAGESIZE chunks.  The address range may be any size provided
 * it is within the physical boundaries.
 */
static int w25q16_write(struct mtd_info *mtd, loff_t to, size_t len,
	size_t *retlen, const u_char *buf)
{
	struct w25q *flash = mtd_to_w25q(mtd);
	loff_t currAddr;
	const u_char *currData;
	u32 sizeInPage;
	int size = len;
	int ret = 0;

	DEBUG(MTD_DEBUG_LEVEL2, "%s: %s %s 0x%08x, len %zd\n",
			flash->mtd.name, __FUNCTION__, "to",
			(u32)to, len);

	if (retlen)
		*retlen = 0;

	/* sanity checks */
	if (!len)
		return(0);

	if (to + len > flash->mtd.size)
		return -EINVAL;

  	down(&flash->lock);

	currAddr = to;
	currData = buf;
	
	/* handle leading un-aligned */
	if(currAddr % FLASH_PAGESIZE != 0 && size > 0)
	{
		sizeInPage = size > (FLASH_PAGESIZE - currAddr % FLASH_PAGESIZE) ? 
			(FLASH_PAGESIZE - currAddr % FLASH_PAGESIZE) : size;
		ret = w25q16_write_page(mtd, currAddr, sizeInPage, currData);
		if(0 != ret)
			goto out;
		
		currData += sizeInPage;
		currAddr += sizeInPage;
		size -= sizeInPage;
		*retlen += sizeInPage;
	}
	
	while(size > FLASH_PAGESIZE)
	{
		ret = w25q16_write_page(mtd, currAddr, FLASH_PAGESIZE, currData);
		if(0 != ret)
			goto out;
		
		currData += FLASH_PAGESIZE;
		currAddr += FLASH_PAGESIZE;
		size -= FLASH_PAGESIZE;
		*retlen += FLASH_PAGESIZE;
	}
	
	if(size)
	{
		ret = w25q16_write_page(mtd, currAddr, size, currData);
		if(0 != ret)
			goto out;
		
		*retlen += size;
	}

out:
	clear_spi_cache();
	up(&flash->lock);

	return ret;
}


/****************************************************************************/

/*
 * SPI device driver setup and teardown
 */

struct flash_info {
	char		*name;
	u8		mfr_id;
	u8		dev_id1;
	u8		dev_id2;
	unsigned int sector_size;
	unsigned int n_sectors;
};

static struct flash_info __devinitdata w25q_data [] = {
	/* REVISIT: fill in JEDEC ids, for parts that have them */
	{ "Winbond SPI W25Q16",	0xef, 0x40, 0x15, 64 * 1024, 32 },
	{ "Winbond SPI W25Q32",	0xef, 0x40, 0x16, 64 * 1024, 64 },
	{ "Winbond SPI W25X32",	0xef, 0x30, 0x16, 64 * 1024, 64 },
	{ "Winbond SPI W25X64",	0xef, 0x30, 0x17, 64 * 1024, 128 },
	{ "MX SPI MX25L3205D",	0xc2, 0x20, 0x16, 64 * 1024, 64 },
	{ "MX SPI MX25L6405D",	0xc2, 0x20, 0x17, 64 * 1024, 128 },
	{ "MX SPI MX25L12805D",	0xc2, 0x20, 0x18, 64 * 1024, 256 }
};

static struct flash_info *w25q_scan(void)
{
	int i;
	unsigned int ids[3];
	unsigned char cmdset[] = {SPI_CMD_JEDEC_ID};

	spi_write_disable();

	// write total size
	WriteRegWord(SPI_REG_CNTL, 0x1);
	// write read length = 3
	WriteRegWord(SPI_REG_TIME_CNTL1, 0x13310501);

	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031001);
	ReadRegWord(SPI_REG_COMMAND);

	for(i = 0; i < sizeof(cmdset) / sizeof(cmdset[0]); i++)
	{
		spi_set_data(cmdset[i]);

		spi_issue_write_cmd();
		
		if(!spi_check_set_status())
		{
			return NULL;
		}
	}

	for(i = 0; i < 3; i++)
	{
		spi_issue_read_cmd();
		if(!spi_check_get_status())
			return NULL;
		spi_get_data(&ids[i]);
	}

	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031020);
	ReadRegWord(SPI_REG_COMMAND);
	WriteRegWord(SPI_REG_COMMAND, 0x04031000);
	ReadRegWord(SPI_REG_COMMAND);
	
	spi_write_disable();

	WriteRegWord(SPI_REG_COMMAND, 0x04031101);
	ReadRegWord(SPI_REG_COMMAND);
	
	for(i = 0; i < sizeof(w25q_data) / sizeof(w25q_data[0]); i++)
	{
		if( ids[0] == w25q_data[i].mfr_id &&
			ids[1] == w25q_data[i].dev_id1 &&
			ids[2] == w25q_data[i].dev_id2)
		{
			printk("Detected %s\n", w25q_data[i].name);
			return &w25q_data[i];
		}
	}
	
	return NULL;
}

/*
 * board specific setup should have ensured the SPI clock used here
 * matches what the READ command supports, at least until this driver
 * understands FAST_READ (for clocks over 25 MHz).
 */
static int __devinit w25q_probe(struct mtd_info *mtd)
{
	struct w25q			*flash = mtd_to_w25q(mtd);
	struct flash_info		*info;
	unsigned	int		i;

	info = w25q_scan();
	if(!info)
		return -ENODEV;

	init_MUTEX(&flash->lock);

	//mtd->name = info->name;
	mtd->name = "physmap-flash.0";
	mtd->type = MTD_NORFLASH;
	mtd->writesize = 1;
	mtd->flags = MTD_CAP_NORFLASH;
	mtd->size = info->sector_size * info->n_sectors;
	mtd->erasesize = info->sector_size;
	mtd->erase = w25q16_erase;
	mtd->read = w25q16_read;
	mtd->write = w25q16_write;

	DEBUG(MTD_DEBUG_LEVEL2,
		"mtd .name = %s, .size = 0x%.8x (%uM) "
			".erasesize = 0x%.8x (%uK) .numeraseregions = %d\n",
		mtd->name,
		mtd->size, mtd->size / (1024*1024),
		mtd->erasesize, mtd->erasesize / 1024,
		mtd->numeraseregions);

	if (mtd->numeraseregions)
	{
		for (i = 0; i < mtd->numeraseregions; i++)
			DEBUG(MTD_DEBUG_LEVEL2,
				"mtd.eraseregions[%d] = { .offset = 0x%.8x, "
				".erasesize = 0x%.8x (%uK), "
				".numblocks = %d }\n",
				i, mtd->eraseregions[i].offset,
				mtd->eraseregions[i].erasesize,
				mtd->eraseregions[i].erasesize / 1024,
				mtd->eraseregions[i].numblocks);
	}

	/* partitions should match sector boundaries; and it may be good to
	 * use readonly partitions for writeprotected sectors (BP2..BP0).
	 */
	if (mtd_has_partitions()) {
		struct mtd_partition	*parts = NULL;
		int			nr_parts = 0;

#ifdef CONFIG_MTD_CMDLINE_PARTS
		static const char *part_probes[] = { "cmdlinepart", NULL, };

		nr_parts = parse_mtd_partitions(&flash->mtd,
				part_probes, &parts, 0);
#endif

		if (nr_parts > 0) {
			flash->partitioned = 1;
			return add_mtd_partitions(&flash->mtd, parts, nr_parts);
		}
	}

	return add_mtd_device(mtd) == 1 ? -ENODEV : 0;
}

#if 0
static int __devexit w25q_remove(struct spi_device *spi)
{
	struct w25q	*flash = dev_get_drvdata(&spi->dev);
	int		status;

	/* Clean up MTD stuff. */
	if (mtd_has_partitions() && flash->partitioned)
		status = del_mtd_partitions(&flash->mtd);
	else
		status = del_mtd_device(&flash->mtd);
	if (status == 0)
		kfree(flash);
	return 0;
}
#endif

#if 0
static struct platform_device w25q16_device = {
    .name       = "w25q16",
    .id     = 1,
    .dev        = {
        .platform_data  = NULL,
    },
    .num_resources  = 0,
};
#endif
#if 0
static struct spi_driver w25q16_driver = {
	.driver = {
		.name	= "w25q16",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
	},
	.probe	= w25q_probe,
	.remove	= __devexit_p(w25q_remove),
};
#endif

struct w25q *flash;

static int w25q16_init(void)
{
	int ret = 0;
	
	flash = (struct w25q *)kzalloc(sizeof(struct w25q), GFP_KERNEL);
	if(!flash)
		return 0;
	
	init_MUTEX(&flash->lock);
	
	ret = w25q_probe(&(flash->mtd));
	
	//ret = platform_device_register
	//ret = spi_register_driver(&w25q16_driver);
	return ret;
}


static void w25q16_exit(void)
{
	if(flash)
		kfree(flash);
	flash = NULL;
	//spi_unregister_driver(&w25q16_driver);
}


module_init(w25q16_init);
module_exit(w25q16_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sam Sheng");
MODULE_DESCRIPTION("MTD SPI driver for Winbond W25Qxx flash chips");
