/*
 *  linux/drivers/mtd/onenand/generic.c
 *
 *  Copyright (c) 2005 Samsung Electronics
 *  Kyungmin Park <kyungmin.park@samsung.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the OneNAND flash for generic boards.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/onenand.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>
#include <asm/delay.h>
/*#include <asm/mach/flash.h>*/
#include <asm/setup.h>
#include <asm/trihidtv/trihidtv.h>

#define DRIVER_NAME	"onenand"


#ifdef CONFIG_MTD_PARTITIONS
//static const char *part_probes[] = { "cmdlinepart", NULL,  };
#endif

extern char arcs_cmdline[];
static char mtd_cmd_line[COMMAND_LINE_SIZE];

/*
 * Define partitions for flash device
 */
#ifdef CONFIG_MTD_PARTITIONS

struct mtd_partition onenand_physical_partition_info[] = {
	{
		name:"mtd0",
		offset:0,
		size:32*1024*1024
	}
};
struct mtd_partition onenand_partition_info[] = {
	{
		name:"Boot(bootloader+kernel+etc.)(6M) On OneNAND",
		offset:0,
		size:6*1024*1024
	},
	{
		name:"JFFS2 file system(26M) On OneNAND",
		offset:6*1024*1024,
		size:26*1024*1024
	}/*,
	{
		name:"Initial Ramdisk image(2M) On OneNAND",
		offset:14*1024*1024,
		size:2*1024*1024
	},
	{
		name:"Working image(16M) On OneNAND",
		offset:16*1024*1024,
		size:16*1024*1024
	}*/
};

#define NUM_PARTITIONS (sizeof(onenand_partition_info)/sizeof(struct mtd_partition))

#endif

#ifdef CONFIG_MTD_ONENAND_CRAMFS_SUPPORT
#ifdef CONFIG_CRAMFS
	extern unsigned int nand_tell_cramfs;
	extern unsigned int cramfs_set_mapping_table(void);
#endif
#endif

struct onenand_info {
	struct mtd_info		mtd;
	struct mtd_partition	*parts;
	struct onenand_chip	onenand;
};

// Add dynamic MTD partitions. [Sam Sheng]
#define MTD_MAX_COUNT 10
static const char MTD_NAME[] = "MTD_NAME=";

static unsigned long long size_parse (char *ptr, char **retptr)
{
	unsigned long long ret = simple_strtoull (ptr, retptr, 16);

	switch (**retptr) {
	case 'G':
	case 'g':
		ret <<= 10;
	case 'M':
	case 'm':
		ret <<= 10;
	case 'K':
	case 'k':
		ret <<= 10;
		(*retptr)++;
	default:
		break;
	}
	return ret;
}

// The format for the command line is as follows:
// mtd_parts	:= partdef[ partdef]
// partdef		:= MTD_NAME=name,start,offset
// name			:= <a-z>[a-z,0-9]
// start		:= standard linux memsize
// offset		:= standard linux memsize
// 
// example: MTD_NAME=KERNEL,0,6M MTD_NAME=CRAMFS,6M,8M MTD_NAME=EXT2,eM,2M
static inline int parse_mtd_cmdline(struct mtd_info *mtd)
{
	char *p, *from = mtd_cmd_line;
	int mtd_count = 0;
	struct mtd_partition mtd_part[MTD_MAX_COUNT];

	memset(mtd_part, 0, sizeof(mtd_part));
	strncpy(mtd_cmd_line, arcs_cmdline, sizeof mtd_cmd_line);
	mtd_cmd_line[sizeof mtd_cmd_line - 1] = 0;
	
	while((from < mtd_cmd_line + COMMAND_LINE_SIZE) && 
			((from = strstr(from, MTD_NAME)) != NULL))
	{
		if(mtd_count >= MTD_MAX_COUNT)
			break;
		if(!(p = strchr(from, ',')))
			break;

		*p = '\0';
		mtd_part[mtd_count].name = from + sizeof(MTD_NAME) - 1;
		mtd_part[mtd_count].offset = size_parse(p + 1, &p);
		mtd_part[mtd_count].size = size_parse(p + 1, &p);
		mtd_count++;
		from = p + 1;
	}

	if(mtd_count > 0){
		printk(KERN_NOTICE
		       "Using partition definition from KERNEL command line parameters\n");
		add_mtd_partitions(mtd, mtd_part, mtd_count);
	}

	return mtd_count;
}

#ifdef CONFIG_MTD_ONENAND_SYNC_MODE
static void check_sync_status(unsigned int s)
{
	int retry = 0;
#define SYNC_STATUS_REG 0x1b00201c
#define SYNC_STATUS_SHIFT 26
#define SYNC_STATUS_MASK 3
#define SYNC_STATUS_MAX_RETRY 100

	while(1)
	{
		unsigned int status;
		unsigned int regVal;

		regVal = ReadRegWord(SYNC_STATUS_REG);
		status = ((regVal >> SYNC_STATUS_SHIFT) & SYNC_STATUS_MASK);
		if(status == s)
		{
			break;
		}
		if(retry++ >= SYNC_STATUS_MAX_RETRY)
		{
			printk("Fatal error: can't recover from sync mode "
				"[0x%08x: 0x%08x]\n", SYNC_STATUS_REG, regVal);
			break;
		}
	}
}

static void hidtv_mmcontrol(struct mtd_info *mtd, int sync_read)
{
	struct onenand_chip *this = mtd->priv;

	if(sync_read)
	{
		this->write_word(0xc0e0, this->base + ONENAND_REG_SYS_CFG1);
		WriteRegWord(0x1b002008, 0x02000002);
	}
	else
	{
		WriteRegWord(0x1b002008, 0x00000002);
		check_sync_status(0);
		this->write_word(0x40c0, this->base + ONENAND_REG_SYS_CFG1);
	}
}
#endif


static int __devinit hidtv_onenand_probe(struct platform_device *dev)
{
	struct onenand_info *info;
	//struct flash_platform_data *pdata = dev->dev.platform_data;
	struct resource *res = dev->resource;
	unsigned long size = res->end - res->start + 1;
	int err;

	info = kzalloc(sizeof(struct onenand_info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	if (!request_mem_region(res->start, size, dev->name)) {
		err = -EBUSY;
		goto out_free_info;
	}

	info->onenand.base = ioremap(res->start, size);
	if (!info->onenand.base) {
		err = -ENOMEM;
		goto out_release_mem_region;
	}

#ifdef CONFIG_MTD_ONENAND_SYNC_MODE
	info->onenand.mmcontrol = hidtv_mmcontrol;
#endif

	//info->onenand.mmcontrol = pdata->mmcontrol;
	info->onenand.irq = platform_get_irq(dev, 0);

	info->mtd.name = dev->dev.bus_id;
	info->mtd.priv = &info->onenand;
	info->mtd.owner = THIS_MODULE;

	if (onenand_scan(&info->mtd, 1)) {
		err = -ENXIO;
		goto out_iounmap;
	}

#if 0
#ifdef CONFIG_MTD_PARTITIONS
	err = parse_mtd_partitions(&info->mtd, part_probes, &info->parts, 0);
	if (err > 0)
		add_mtd_partitions(&info->mtd, info->parts, err);
	else if (err < 0 && pdata->parts)
		add_mtd_partitions(&info->mtd, pdata->parts, pdata->nr_parts);
	else
#endif
		err = add_mtd_device(&info->mtd);
#endif

	dev_set_drvdata(&dev->dev, info);

	/*
	 * The whole nand flash...
	 */
	/* Register the partition */
	onenand_physical_partition_info[0].size = info->mtd.size;
	add_mtd_partitions(&info->mtd, onenand_physical_partition_info, 1);

	if (!parse_mtd_cmdline(&info->mtd))
	{
		/* Register the default partitions */
		add_mtd_partitions(&info->mtd, onenand_partition_info, NUM_PARTITIONS);
	}

#ifdef CONFIG_MTD_ONENAND_CRAMFS_SUPPORT
#ifdef CONFIG_CRAMFS
	/*
	 * Always put this to last!
	 */
	nand_tell_cramfs = 1;
	cramfs_set_mapping_table();
#endif
#endif

	return 0;

out_iounmap:
	iounmap(info->onenand.base);
out_release_mem_region:
	release_mem_region(res->start, size);
out_free_info:
	kfree(info);

	return err;
}

static int __devexit hidtv_onenand_remove(struct platform_device *dev)
{
	struct onenand_info *info = dev_get_drvdata(&dev->dev);
	struct resource *res = dev->resource;
	unsigned long size = res->end - res->start + 1;

	dev_set_drvdata(&dev->dev, NULL);

	if (info) {
		if (info->parts)
			del_mtd_partitions(&info->mtd);
		else
			del_mtd_device(&info->mtd);

		onenand_release(&info->mtd);
		release_mem_region(res->start, size);
		iounmap(info->onenand.base);
		kfree(info);
	}
	
#ifdef CONFIG_MTD_ONENAND_CRAMFS_SUPPORT
#ifdef CONFIG_CRAMFS
	    printk("%s:Tell CRAMFS that nand is over.\n",__FUNCTION__);
		nand_tell_cramfs = 0;
#endif
#endif

	return 0;
}

static struct platform_driver onenand_flash_driver = {
	.probe		= hidtv_onenand_probe,
	.remove		= hidtv_onenand_remove,
#ifdef CONFIG_PM
	.suspend	= NULL,
	.resume		= NULL,
	.shutdown	= NULL,
#endif
	.driver		= {
		.name	= "onenand-flash",
	},
};

static struct resource onenand_flash_resource = {
#if defined(CONFIG_MTD_ONENAND_PHYSMAP_START) && defined(CONFIG_MTD_ONENAND_PHYSMAP_LEN)
	.start		= CONFIG_MTD_ONENAND_PHYSMAP_START,
#else
	.start		= 0x1c000000,
#endif
#if defined(CONFIG_MTD_ONENAND_PHYSMAP_START) && defined(CONFIG_MTD_ONENAND_PHYSMAP_LEN)
	.end		= CONFIG_MTD_ONENAND_PHYSMAP_START + CONFIG_MTD_ONENAND_PHYSMAP_LEN - 1,
#else
	.end		= 0x1c000000 + 0x1000000 - 1,
#endif
	.flags		= IORESOURCE_MEM,
};

static struct platform_device onenand_flash_device = {
	.name		= "onenand-flash",
	.id		= 1,
	.dev		= {
		.platform_data	= NULL,
	},
	.num_resources	= 1,
	.resource	= &onenand_flash_resource,
};

MODULE_ALIAS(DRIVER_NAME);

static int __init hidtv_onenand_init(void)
{
	int err;

	err = platform_device_register(&onenand_flash_device);
	if(err == 0)
		err = platform_driver_register(&onenand_flash_driver);

	return err;
}

static void __exit hidtv_onenand_exit(void)
{
	platform_driver_unregister(&onenand_flash_driver);
	platform_device_unregister(&onenand_flash_device);
}

module_init(hidtv_onenand_init);
module_exit(hidtv_onenand_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kyungmin Park <kyungmin.park@samsung.com>");
MODULE_DESCRIPTION("Glue layer for OneNAND flash on Trident HiDTV boards");
