/*
 *  nandwrite.c
 *
 *  Copyright (C) 2000 Steven J. Hill (sjhill@realitydiluted.com)
 *   		  2003 Thomas Gleixner (tglx@linutronix.de)
 *
 * $Id: nandwrite.c,v 1.1 2007/09/19 08:43:38 samsheng Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Overview:
 *   This utility writes a binary image directly to a NAND flash
 *   chip or NAND chips contained in DoC devices. This is the
 *   "inverse operation" of nanddump.
 *
 * tglx: Major rewrite to handle bad blocks, write data with or without ECC
 *	 write oob data only on request
 *
 * Bug/ToDo:
 */

#define _GNU_SOURCE
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <getopt.h>

#include <asm/types.h>
#include "mtd/mtd-user.h"

#define PROGRAM "nandwrite"
#define VERSION "1.2"

#define BUILTIN_OOB

/*
 * Buffer array used for writing data
 */
unsigned char *writebuf;
unsigned char oobbuf[32];

// oob layouts to pass into the kernel as default
struct nand_oobinfo none_oobinfo = { 
	.useecc = MTD_NANDECC_OFF,
};

struct nand_oobinfo jffs2_oobinfo = { 
	.useecc = MTD_NANDECC_PLACE,
	.eccbytes = 6,
	.eccpos = { 0, 1, 2, 3, 6, 7 }
};

struct nand_oobinfo yaffs_oobinfo = { 
	.useecc = MTD_NANDECC_PLACE,
	.eccbytes = 6,
	.eccpos = { 8, 9, 10, 13, 14, 15}
};

struct nand_oobinfo autoplace_oobinfo = {
	.useecc = MTD_NANDECC_AUTOPLACE
};

void display_help (void)
{
	printf("Usage: nandwrite [OPTION] MTD_DEVICE INPUTFILE\n"
	       "Writes to the specified MTD device.\n"
	       "\n"
	       "  -a, --autoplace  	Use auto oob layout\n"
	       "  -j, --jffs2  	 	force jffs2 oob layout\n"
	       "  -y, --yaffs  	 	force yaffs oob layout\n"
	       "  -n, --noecc		write without ecc\n"
		   "  -k addr, --skip-start=addr skip start address (default is 0xffffffff)\n"
		   "  -l len, --skip-length=len skip length (default is 0)\n"
	       "  -o, --oob    	 	use built-in oob data(0x19 0x85 0x20)\n"
	       "  -s addr, --start=addr set start address (default is 0)\n"
		   "  -u addr, --upper=addr set upper boundary (default is 0xffffffff)\n"
	       "  -p, --pad             pad to page size\n"
	       "  -q, --quiet    	don't display progress messages\n"
	       "      --help     	display this help and exit\n"
	       "      --version  	output version information and exit\n");
	exit(0);
}

void display_version (void)
{
	printf(PROGRAM " " VERSION "\n"
	       "\n"
	       "Copyright (C) 2003 Thomas Gleixner \n"
	       "\n"
	       PROGRAM " comes with NO WARRANTY\n"
	       "to the extent permitted by law.\n"
	       "\n"
	       "You may redistribute copies of " PROGRAM "\n"
	       "under the terms of the GNU General Public Licence.\n"
	       "See the file `COPYING' for more information.\n");
	exit(0);
}

char 	*mtd_device, *img;
int 	mtdoffset = 0;
int 	quiet = 0;
int	writeoob = 0;
int	autoplace = 0;
int	forcejffs2 = 0;
int	forceyaffs = 0;
int	noecc = 0;
int	pad = 0;
unsigned int skip_start = 0xffffffff;
unsigned int skip_length = 0;
unsigned int upper_boundary = 0xffffffff;

unsigned char oob_clean_marker[32] = {
	0xff, 0xff, 0x19, 0x85, 0x20, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

void process_options (int argc, char *argv[])
{
	int error = 0;

	for (;;) {
		int option_index = 0;
		static const char *short_options = "os:ajynqpk:l:u:";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 0},
			{"version", no_argument, 0, 0},
			{"oob", no_argument, 0, 'o'},
			{"start", required_argument, 0, 's'},
			{"autoplace", no_argument, 0, 'a'},
			{"jffs2", no_argument, 0, 'j'},
			{"yaffs", no_argument, 0, 'y'},
			{"noecc", no_argument, 0, 'n'},
			{"quiet", no_argument, 0, 'q'},
			{"pad", no_argument, 0, 'p'},
			{"skip-start", required_argument, 0, 'k'},
			{"skip-length", required_argument, 0, 'l'},
			{"upper", required_argument, 0, 'u'},
			{0, 0, 0, 0},
		};

		int c = getopt_long(argc, argv, short_options,
				    long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 0:
			switch (option_index) {
			case 0:
				display_help();
				break;
			case 1:
				display_version();
				break;
			}
			break;
		case 'q':
			quiet = 1;
			break;
		case 'a':
			autoplace = 1;
			break;
		case 'j':
			forcejffs2 = 1;
			break;
		case 'y':
			forceyaffs = 1;
			break;
		case 'n':
			noecc = 1;
			break;
		case 'o':
			writeoob = 1;
			break;
		case 'p':
			pad = 1;
			break;
		case 's':
			mtdoffset = strtoul(optarg, NULL, 0);
			break;
		case 'k':
			skip_start = strtoul(optarg, NULL, 0);
			break;
		case 'l':
			skip_length = strtoul(optarg, NULL, 0);
			break;
		case 'u':
			upper_boundary = strtoul(optarg, NULL, 0);
			break;
		case '?':
			error = 1;
			break;
		}
	}
	
	if ((argc - optind) != 2 || error) 
		display_help ();
	
	mtd_device = argv[optind++];
	img = argv[optind];
}

/*
 * Main program
 */
int main(int argc, char **argv)
{
	int cnt, fd, ifd, imglen, pagelen, blockstart = -1;
	struct mtd_info_user meminfo;
	struct mtd_oob_buf oob;
	loff_t offs;
	int ret, readlen;

	process_options(argc, argv);

#ifndef BUILTIN_OOB
	if (pad && writeoob) {
		fprintf(stderr, "Can't pad when oob data is present.\n");
		exit(1);
	}
#endif

	/* Open the device */
	if ((fd = open(mtd_device, O_RDWR)) == -1) {
		perror("open flash");
		exit(1);
	}

	/* Fill in MTD device capability structure */  
	if (ioctl(fd, MEMGETINFO, &meminfo) != 0) {
		perror("MEMGETINFO");
		close(fd);
		exit(1);
	}

	/* Make sure device page sizes are valid */
	if (!(meminfo.oobsize == 16 && meminfo.writesize == 512) &&
	    !(meminfo.oobsize == 8 && meminfo.writesize == 256) && 
	    !(meminfo.oobsize == 32 && meminfo.writesize == 1024) && 
	    !(meminfo.oobsize == 64 && meminfo.writesize == 2048)) {
		fprintf(stderr, "Unknown flash (not normal NAND)\n");
		close(fd);
		exit(1);
	}
	
	// write without ecc ?
	if (noecc) {
		if (ioctl (fd, MEMSETOOBSEL, &none_oobinfo) != 0) {
			perror ("MEMSETOOBSEL");
			close (fd);
			exit (1);
		} 
	}

	// autoplace ECC ?
	if (autoplace) {
		if (ioctl (fd, MEMSETOOBSEL, &autoplace_oobinfo) != 0) {
			perror ("MEMSETOOBSEL");
			close (fd);
			exit (1);
		} 
	}

	// force oob layout for jffs2 or yaffs ?
	if (forcejffs2 || forceyaffs) {
		struct nand_oobinfo *oobsel = forcejffs2 ? &jffs2_oobinfo : &yaffs_oobinfo;
		
		if (forceyaffs && meminfo.oobsize == 8) {
    			if (forceyaffs) {
				fprintf (stderr, "YAFSS cannot operate on 256 Byte page size");
				close (fd);
				exit (1);
			}	
			/* Adjust number of ecc bytes */	
			jffs2_oobinfo.eccbytes = 3;	
		}
		
		if (ioctl (fd, MEMSETOOBSEL, oobsel) != 0) {
			perror ("MEMSETOOBSEL");
			close (fd);
			exit (1);
		} 
	}

	oob.length = meminfo.oobsize;
	oob.ptr = oobbuf;

	/* Open the input file */
	if ((ifd = open(img, O_RDONLY)) == -1) {
		perror("open input file");
		close(fd);
		close(ifd);
		exit(1);
	}

	// get image length
   	imglen = lseek(ifd, 0, SEEK_END);
	lseek (ifd, 0, SEEK_SET);

#ifndef BUILTIN_OOB	
	pagelen = meminfo.writesize + ((writeoob == 1) ? meminfo.oobsize : 0);
#else
	pagelen = meminfo.writesize;
#endif
	
#ifndef BUILTIN_OOB
	// Check, if file is pagealigned
	if ((!pad) && ((imglen % pagelen) != 0)) {
		printf("Input file is not page aligned img len: %#010x page len: %#010x\n", imglen, pagelen);
		close (fd);
		close (ifd);
		exit (1);
	}
#endif
	
	// Check, if length fits into device
	if ( ((imglen / pagelen) * meminfo.writesize) > (meminfo.size - mtdoffset)) {
		perror ("Input file does not fit into device");
		close (fd);
		close (ifd);
		exit (1);
	}

	writebuf = malloc(meminfo.writesize);
	if (writebuf == NULL)
	{
		perror ("Failed in memory alloc");
		close (fd);
		close (ifd);
		exit (1);
	}
	
	/* Get data from input and write to the device */
	while (imglen && (mtdoffset < meminfo.size)) {
		// new eraseblock , check for bad block
		if (blockstart != (mtdoffset & (~meminfo.erasesize + 1))) {
			blockstart = mtdoffset & (~meminfo.erasesize + 1);
			offs = blockstart;
			if (!quiet)
				fprintf (stdout, "Writing data to block 0x%08x\n", blockstart);
			if ((ret = ioctl(fd, MEMGETBADBLOCK, &offs)) < 0) {
#if 0			/* seems it's NOR flash, we won't quit here */
				perror("ioctl(MEMGETBADBLOCK)");
				close (ifd);
				close (fd);
				free(writebuf);
				exit (1);
#endif
			}
			if (ret == 1) {
				if (!quiet)
					fprintf (stderr, "Bad block at %x, will be skipped\n", blockstart);
				mtdoffset = blockstart + meminfo.erasesize;
				continue;					
			}
		}

		if(mtdoffset >= upper_boundary)
		{
			fprintf(stderr, "Exceeds upper limit 0x%x, too many bad blocks?\n", 
					upper_boundary);
			close (ifd);
			close (fd);
			free(writebuf);
			exit (1);
		}

		readlen = meminfo.writesize;
		if (pad && (imglen < readlen))
		{
			readlen = imglen;
			memset(writebuf + readlen, 0xff, meminfo.writesize - readlen);
		}

		/* Read Page Data from input file */
		if ((cnt = read(ifd, writebuf, readlen)) != readlen) {
			if (readlen == 0)	// EOF
				break;
			perror ("File I/O error on input file");
			close (fd);
			close (ifd);
			free(writebuf);
			exit (1);						
		}	

		if(skip_start <= mtdoffset && skip_length > 0)
		{
			/* we skip to write this block */
			if(mtdoffset % meminfo.erasesize == 0)
				printf("Skip writing block 0x%x size 0x%x\n", mtdoffset, 
						skip_length > meminfo.erasesize ? meminfo.erasesize : skip_length);
			skip_length -= readlen;
			imglen -= readlen;
			mtdoffset += meminfo.writesize;
			continue;
		}
		
		if (writeoob && mtdoffset % meminfo.erasesize == 0) {
#ifndef BUILTIN_OOB 
			/* Read OOB data from input file, exit on failure */
			if ((cnt = read(ifd, oobbuf, meminfo.oobsize)) != meminfo.oobsize) {
				perror ("File I/O error on input file");
				close (fd);
				close (ifd);
				free(writebuf);
				exit (1);						
			}
#else
			memcpy(oobbuf, oob_clean_marker, sizeof(oobbuf));
#endif
			/* Write OOB data first, as ecc will be placed in there*/
			oob.start = mtdoffset;
			if (ioctl(fd, MEMWRITEOOB, &oob) != 0) {
#if 0			/* seems it's NOR flash, we won't quit here */
				perror ("ioctl(MEMWRITEOOB)");
				close (fd);
				close (ifd);
				free(writebuf);
				exit (1);
#endif
			}
#ifndef BUILTIN_OOB
			imglen -= meminfo.oobsize;
#endif
		}
		
		/* Write out the Page data */
		if (pwrite(fd, writebuf, meminfo.writesize, mtdoffset) != meminfo.writesize) {
			perror ("pwrite");
			close (fd);
			close (ifd);
			free(writebuf);
			exit (1);
		}

		imglen -= readlen;
		mtdoffset += meminfo.writesize;
	}

	/* Close the output file and MTD device */
	close(fd);
	close(ifd);
	free(writebuf);

	if (imglen > 0) {
		perror ("Data did not fit into device, due to bad blocks\n");
		exit (1);
	}

	sync();

	/* Return happy */
	return 0;
}
