/*
 *  nandwrite.c
 *
 *  Copyright (C) 2000 Steven J. Hill (sjhill@realitydiluted.com)
 *   		  2003 Thomas Gleixner (tglx@linutronix.de)
 *
 * $Id: norwrite2.c,v 1.5 2008/04/11 10:53:43 samsheng Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Overview:
 *   This utility writes a binary image directly to a NAND flash
 *   chip or NAND chips contained in DoC devices. This is the
 *   "inverse operation" of nanddump.
 *
 * tglx: Major rewrite to handle bad blocks, write data with or without ECC
 *	 write oob data only on request
 *
 * Bug/ToDo:
 */

#define _GNU_SOURCE
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <getopt.h>
#include <limits.h>

#include <asm/types.h>
#include "mtd/mtd-user.h"

#define PROGRAM "norwrite"
#define VERSION "1.2"

/*
 * Buffer array used for writing data
 */
unsigned char *writebuf;

void display_help (void)
{
	printf("Usage: norwrite [OPTION] MTD_DEVICE INPUTFILE\n"
	       "Writes to the specified MTD device.\n"
	       "\n"
	       "  -k addr, --skip-start=addr skip start address (default is 0xffffffff)\n"
	       "  -l len, --skip-length=len skip length (default is 0)\n"
	       "  -s addr, --start=addr set start address (default is 0)\n"
		   "  -u addr, --upper=addr set upper boundary (default is 0xffffffff)\n"
	       "  -p, --pad             pad to page size\n"
		   "  -f, --fifo		fifo updating indicator\n"
	       "  -q, --quiet    	don't display progress messages\n"
	       "      --help     	display this help and exit\n"
	       "      --version  	output version information and exit\n");
	exit(0);
}

void display_version (void)
{
	printf(PROGRAM " " VERSION "\n"
	       "\n"
	       "Copyright (C) 2003 Thomas Gleixner \n"
	       "\n"
	       PROGRAM " comes with NO WARRANTY\n"
	       "to the extent permitted by law.\n"
	       "\n"
	       "You may redistribute copies of " PROGRAM "\n"
	       "under the terms of the GNU General Public Licence.\n"
	       "See the file `COPYING' for more information.\n");
	exit(0);
}

char 	*mtd_device, *img;
int 	mtdoffset = 0;
int 	quiet = 0;
int	pad = 0;
unsigned int skip_start = 0xffffffff;
unsigned int skip_length = 0;
unsigned int upper_boundary = 0xffffffff;
char fifo_path[PATH_MAX] = "upd_stat";

void process_options (int argc, char *argv[])
{
	int error = 0;

	for (;;) {
		int option_index = 0;
		static const char *short_options = "s:qpk:l:u:f:";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 0},
			{"version", no_argument, 0, 0},
			{"start", required_argument, 0, 's'},
			{"quiet", no_argument, 0, 'q'},
			{"pad", no_argument, 0, 'p'},
			{"skip-start", required_argument, 0, 'k'},
			{"skip-length", required_argument, 0, 'l'},
			{"upper", required_argument, 0, 'u'},
			{"fifo", required_argument, 0, 'f'},
			{0, 0, 0, 0},
		};

		int c = getopt_long(argc, argv, short_options,
				    long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 0:
			switch (option_index) {
			case 0:
				display_help();
				break;
			case 1:
				display_version();
				break;
			}
			break;
		case 'q':
			quiet = 1;
			break;
		case 'p':
			pad = 1;
			break;
		case 's':
			mtdoffset = strtoul(optarg, NULL, 0);
			break;
		case 'k':
			skip_start = strtoul(optarg, NULL, 0);
			break;
		case 'l':
			skip_length = strtoul(optarg, NULL, 0);
			break;
		case 'u':
			upper_boundary = strtoul(optarg, NULL, 0);
			break;
		case 'f':
			strcpy(fifo_path, optarg);
			break;
		case '?':
			error = 1;
			break;
		}
	}
	
	if ((argc - optind) != 2 || error) 
		display_help ();
	
	mtd_device = argv[optind++];
	img = argv[optind];
}

static void
log_updating_status(int fd, unsigned int offset)
{
	char str[32];

	sprintf(str, "[0x%x]", offset);
	write(fd, str, strlen(str));
}

/*
 * Main program
 */
int main(int argc, char **argv)
{
	int cnt, fd, ifd, imglen, pagelen, blockstart = -1;
	struct mtd_info_user meminfo;
	loff_t offs;
	int readlen;
	erase_info_t erase;
	int fifo_fd;

	process_options(argc, argv);

	fifo_fd = open(fifo_path, O_WRONLY | O_NONBLOCK);
	if(fifo_fd < 0)
	{
		fprintf(stderr, 
				"Warning, can't open updating status indicating fifo\n");
	}
	
	/* Open the device */
	if ((fd = open(mtd_device, O_RDWR)) == -1) {
		perror("open flash");
		exit(1);
	}

	/* Fill in MTD device capability structure */  
	if (ioctl(fd, MEMGETINFO, &meminfo) != 0) {
		perror("MEMGETINFO");
		close(fd);
		exit(1);
	}

	if (meminfo.oobsize != 0)
	{
		fprintf(stderr, "it's not a nor flash\n");
		exit(1);
	}

	if(mtdoffset % meminfo.erasesize != 0)
	{
		fprintf(stderr, "Partition start address 0x%08x "
				"is not block size 0x%x aligned\n",
				mtdoffset, meminfo.erasesize);
		exit(1);
	}

	/* Open the input file */
	if ((ifd = open(img, O_RDONLY)) == -1) {
		perror("open input file");
		close(fd);
		exit(1);
	}

	// get image length
   	imglen = lseek(ifd, 0, SEEK_END);
	lseek (ifd, 0, SEEK_SET);

	pagelen = meminfo.erasesize;
	erase.length = meminfo.erasesize;
	
	// Check, if length fits into device
	if ( imglen > (meminfo.size - mtdoffset)) {
		perror ("Input file does not fit into device");
		close (fd);
		close (ifd);
		exit (1);
	}

	writebuf = malloc(meminfo.erasesize);
	if (writebuf == NULL)
	{
		perror ("Failed in memory alloc");
		close (fd);
		close (ifd);
		exit (1);
	}
	
	/* Get data from input and write to the device */
	while (imglen && (mtdoffset < meminfo.size)) {
		// new eraseblock , check for bad block
		if (blockstart != (mtdoffset & (~meminfo.erasesize + 1))) {
			blockstart = mtdoffset & (~meminfo.erasesize + 1);
			offs = blockstart;
		}

		readlen = meminfo.erasesize;
		if (pad && (imglen < readlen))
		{
			readlen = imglen;
			memset(writebuf + readlen, 0xff, meminfo.erasesize - readlen);
		}

		/* Read Page Data from input file */
		if ((cnt = read(ifd, writebuf, readlen)) != readlen) {
			if (readlen == 0)	// EOF
				break;
			perror ("File I/O error on input file");
			close (fd);
			close (ifd);
			free(writebuf);
			exit (1);						
		}	

		log_updating_status(fifo_fd, mtdoffset);

		if(skip_start <= mtdoffset && skip_length > 0)
		{
			/* we skip to write this block */
			if(mtdoffset % meminfo.erasesize == 0)
				printf("Skip writing block 0x%x size 0x%x\n", mtdoffset, 
						skip_length > meminfo.erasesize ? meminfo.erasesize : skip_length);
			skip_length -= readlen;
			imglen -= readlen;
			mtdoffset += meminfo.erasesize;
			continue;
		}
		
		erase.start = mtdoffset;
		/*
		if (!quiet)
			fprintf(stdout, "Unlocking block 0x%08x\n", blockstart);
		if (ioctl(fd,MEMUNLOCK,&erase) != 0)
		{
			perror("\nMTD Unlock failure");
		}
		 */
		/*
		if (!quiet)
			fprintf(stdout, "Erasing block 0x%08x\n", blockstart);
		 */
		if (ioctl(fd,MEMERASE,&erase) != 0)
		{
			perror("\nMTD Erase failure");
			close(fd);
			exit (1);
		}
		if (!quiet)
			fprintf (stdout, "Writing data to block 0x%08x\n", blockstart);
		
		/* Write out the Page data */
		if (pwrite(fd, writebuf, readlen, mtdoffset) != readlen) {
			perror ("pwrite");
			close (fd);
			close (ifd);
			free(writebuf);
			exit (1);
		}

		imglen -= readlen;
		mtdoffset += meminfo.erasesize;
	}
	
	mtdoffset += meminfo.erasesize - 1;
	mtdoffset &= (~meminfo.erasesize + 1);
	while(upper_boundary != 0xffffffff && mtdoffset < upper_boundary)
	{
		int ret;
		if ((ret = ioctl(fd, MEMGETBADBLOCK, &mtdoffset)) < 0) {
#if 0			/* seems it's NOR flash, we won't quit here */
			perror("ioctl(MEMGETBADBLOCK)");
			close (ifd);
			close (fd);
			free(writebuf);
			exit (1);
#endif
		}
		if (ret == 1) {
			if (!quiet)
				fprintf (stderr, "Bad block at %x, will be skipped\n", mtdoffset);
			mtdoffset += meminfo.erasesize;
			continue;
		}
		
		log_updating_status(fifo_fd, mtdoffset);

		if(skip_start <= mtdoffset && skip_length > 0)
		{
			/* we skip to write this block */
			if(mtdoffset % meminfo.erasesize == 0)
				printf("Skip writing block 0x%x size 0x%x\n", mtdoffset, 
						skip_length > meminfo.erasesize ? meminfo.erasesize : skip_length);
			skip_length -= meminfo.erasesize;
			mtdoffset += meminfo.erasesize;
			continue;
		}
		
		erase.start = mtdoffset;
		/*
		if (ioctl(fd,MEMUNLOCK,&erase) != 0)
		{
			perror("\nMTD Unlock failure");
		}
		 */
		if (ioctl(fd,MEMERASE,&erase) != 0)
		{
			perror("\nMTD Erase failure");
			close(fd);
			exit (1);
		}
		
		mtdoffset += meminfo.erasesize;
	}

	/* Close the output file and MTD device */
	close(fd);
	close(ifd);
	if(fifo_fd >= 0)
		close(fifo_fd);
	free(writebuf);

	if (imglen > 0) {
		perror ("Data did not fit into device, due to bad blocks\n");
		exit (1);
	}

	sync();

	/* Return happy */
	return 0;
}
